#!/usr/bin/env python3
import argparse
import json
import os
import shutil
import sys
import requests
import glob
from alphaplan_keys import ApKeys, ApSubKeys, get_ap_dict


def ap_send_json(jsonobj):
    """Sends the generated files to the Alphaplan webservice"""
    url = os.environ.get("AP_WEBSERVICE_URL")
    usr = os.environ.get("AP_WEBSERVICE_USR")
    pw = os.environ.get("AP_WEBSERVICE_PW")
    cert = os.environ.get("AP_WEBSERVICE_CERT")
    # Write TLS certificate to file
    cert_file = "GarzFrickeGmbH-CA.cer"
    with open(cert_file, "w", encoding="utf-8") as f:
        f.write(cert)
    # Send data
    msg = requests.post(url, json=jsonobj, auth=(usr, pw), verify=cert_file, timeout=10)
    os.remove(cert_file)
    msg_json = msg.json()
    if msg_json["Status"] != "Ok":
        sys.exit("ERROR: AlphaPlan webservice post request failed")
    print("AlphaPlan webservice response: {}".format(msg_json["Meldung"]))


def ap_id_generator():
    """Returns a unique ids for each new article"""
    if not hasattr(ap_id_generator, "value"):
        ap_id_generator.value = 0
    ap_id_generator.value += 1
    return "Neu{}".format(ap_id_generator.value)


def new_ap_article(
    warengruppe,
    matchcode,
    bezeichnung,
    langtext,
    attributeset,
    attribute,
    stueckliste=None,
):
    """Creates a dict/list structure for a new AlphaPlan article"""
    position = {}
    for idx in range(len(attribute)):
        if attribute[idx]:
            position["Attribut{:02d}".format(idx + 1)] = attribute[idx]

    data = {
        "Artikel": {
            "ID": ap_id_generator(),
            "Warengruppe": warengruppe,
            "MatchCode": matchcode,
            "Bezeichnung": bezeichnung,
            "Langtext": langtext,
            "Attribute": {"AttributeSet": attributeset, "Position": [position]},
        }
    }

    if stueckliste:
        sub_articles = {
            "Stueckliste": {"Info": "Infofeld in Stueckliste", "Position": []}
        }
        for article in stueckliste:
            sub_articles["Stueckliste"]["Position"].append(article)
        data["Artikel"].update(sub_articles)

    return data


def known_ap_article(artikelnummer):
    """Creates a dict for a known AlphaPlan FWR article"""
    data = {"Artikel": {"ID": artikelnummer, "Artikelnummer": artikelnummer}}
    return data


def generate_ap_subarticle(
    artifacts, ap_key, output_dir, machine, machine_ap, release_name_ap, md5sums
):
    """Create an new AlphaPlan FWR subarticle which is added to a part list"""
    # Generate a dict for all artifacts that should be included in the FWR
    ap_dict = get_ap_dict(machine, machine_ap, release_name_ap)
    # Match artifacts and get path and md5sum
    for artifact in artifacts:
        # Compare the artifact with the match from ap_key
        if artifact.casefold().endswith(ap_dict[ap_key][ApSubKeys.MATCH].casefold()):
            path = os.path.join(output_dir, machine, artifact)
            path = os.path.join(*(path.split(os.path.sep)[2:]))
            # Distinguish between FNGSystem and Yocto path on drive Z:
            if "fngsystem".casefold() in output_dir.casefold():
                path = (
                    "Z:\\Development\\SoftwareStore\\Flash-N-Go\\"
                    "FNGSystem\\{}".format(path.replace("/", "\\"))
                )
            else:
                path = "Z:\\Development\\SoftwareStore\\Linux-Yocto\\{}".format(
                    path.replace("/", "\\")
                )
            md5sum = md5sums[artifact]
    if "path" not in locals():
        sys.exit("ERROR: Can not find key:{} in artifacts".format(ap_key))

    # AlphaPlan specific entries for attribute set: "Firmware, Bestandteil..."
    # Field1: type, Field2: path to artifact, Field3: md5sum
    attribute = [ap_dict[ap_key][ApSubKeys.TYP], path, md5sum]

    return new_ap_article(
        "FWR",
        ap_dict[ap_key][ApSubKeys.MATCHCODE],
        ap_dict[ap_key][ApSubKeys.BEZEICHNUNG],
        ap_dict[ap_key][ApSubKeys.LANGTEXT],
        ap_dict[ap_key][ApSubKeys.ATTRIBUTESET],
        attribute,
    )


def generate_fwr_articles(
    output_dir, outlocal_dir, machine, release_name_local, artifacts_all, md5sums
):
    """Main function to generate the FWR articles for FNGSystem and Yocto"""
    # Modify name strings for AlphaPlan
    machine_ap = machine.upper()
    machine_ap = machine_ap.replace("IMX", "i.MX")
    machine_ap = machine_ap.replace("GUF", "")
    release_name_ap = release_name_local.replace("Yocto-", "Yocto ")
    attb_set_major = "Softwarepaket 32-bit"

    stueckliste = []
    if "fngsystem".casefold() in release_name_local.casefold():
        # FNGSystem Release
        # Set artifact keys
        subarticles = [
            ApKeys.FNGSYS_INIT,
            ApKeys.FNGSYS_UPDATE,
            ApKeys.FNGSYS_FS,
            ApKeys.FNGSYS_CHECKSUM,
        ]
        for key in subarticles:
            stueckliste.append(
                generate_ap_subarticle(
                    artifacts_all,
                    key,
                    outlocal_dir,
                    machine,
                    machine_ap,
                    release_name_ap,
                    md5sums,
                )
            )
        # Define attribute fields
        attribute = [None, None, "FNG-SYSTEM"]
        # Generate dict/list structure
        data = new_ap_article(
            "FWR",
            "SW-Paket",
            "{} {}".format(machine_ap, release_name_ap),
            "{}\n{}".format(release_name_ap, machine_ap),
            attb_set_major,
            attribute,
            stueckliste=stueckliste,
        )
        # If imx-boot.tar.gz or deprecated imx-boot is in artifacts_all (imx8mX)i
        # add an uboot FWR package
        if "imx-boot.tar.gz" in artifacts_all or "imx-boot" in artifacts_all:
            if "imx-boot.tar.gz" in artifacts_all:
                subarticles_uboot = [
                    ApKeys.FNGSYS_UBOOT_UPDATE,
                    ApKeys.FNGSYS_UBOOT_IMAGETAR,
                ]
            # this is not needed if the above tar.gz is used.
            elif "imx-boot" in artifacts_all:
                subarticles_uboot = [
                    ApKeys.FNGSYS_UBOOT_UPDATE,
                    ApKeys.FNGSYS_UBOOT_IMAGE,
                    ApKeys.FNGSYS_UBOOT_CHECKSUM,
                ]
            stueckliste_uboot = []
            for key in subarticles_uboot:
                stueckliste_uboot.append(
                    generate_ap_subarticle(
                        artifacts_all,
                        key,
                        outlocal_dir,
                        machine,
                        machine_ap,
                        release_name_ap,
                        md5sums,
                    )
                )

            # At the moment there are no attributes specified for uboot FWR
            attribute_uboot = [None, None, "UBOOT"]
            data_uboot = new_ap_article(
                "FWR",
                "SW-Paket",
                "{} U-Boot {}".format(machine_ap, release_name_ap),
                "{}\n{}".format(release_name_ap, machine_ap),
                attb_set_major,
                attribute_uboot,
                stueckliste=stueckliste_uboot,
            )
    else:
        # Same process for a yocto release
        subarticles = [ApKeys.YOCTO_PKG_PY, ApKeys.YOCTO_FNG_INSTALL, ApKeys.YOCTO_FS]
        for key in subarticles:
            stueckliste.append(
                generate_ap_subarticle(
                    artifacts_all,
                    key,
                    outlocal_dir,
                    machine,
                    machine_ap,
                    release_name_ap,
                    md5sums,
                )
            )
        attribute = [None, None, None, "YOCTO"]
        data = new_ap_article(
            "FWR",
            "SW-Paket",
            "{} {}".format(machine_ap, release_name_ap),
            "{}\n{}".format(release_name_ap, machine_ap),
            attb_set_major,
            attribute,
            stueckliste=stueckliste,
        )

    # Create an additional import file for uboot
    if "data_uboot" in locals():
        print(
            "Generate {} U-Boot {} AlphaPlan FWR articles".format(
                machine_ap, release_name_ap
            )
        )
        jsonfile_uboot_name = "alphaplan-import-uboot-{}.json".format(machine)
        jsonfile_uboot_local = os.path.join(outlocal_dir, machine, jsonfile_uboot_name)
        with open(jsonfile_uboot_local, "w", encoding="utf-8") as jsonfile:
            json.dump(data_uboot, jsonfile, indent=4)

        if output_dir is not None:
            jsonfile_uboot_output = os.path.join(
                output_dir, machine, jsonfile_uboot_name
            )
            shutil.copyfile(
                jsonfile_uboot_local, jsonfile_uboot_output, follow_symlinks=True
            )
        # Send data object to AlphaPlan webservice
        ap_send_json(data_uboot)

    print("Generate {} {} AlphaPlan FWR articles".format(machine_ap, release_name_ap))
    # Generate a json file from the dict/list structure for debugging
    jsonfile_name = "alphaplan-import-{}.json".format(machine)
    jsonfile_local = os.path.join(outlocal_dir, machine, jsonfile_name)
    with open(jsonfile_local, "w", encoding="utf-8") as jsonfile:
        json.dump(data, jsonfile, indent=4)

    if output_dir is not None:
        # Copy file to release folder as the atrifacts have already been copied
        jsonfile_output = os.path.join(output_dir, machine, jsonfile_name)
        shutil.copyfile(jsonfile_local, jsonfile_output, follow_symlinks=True)
    # Send data object to AlphaPlan webservice
    ap_send_json(data)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--images-dir",
        help="""Yocto images directory""",
        dest="images_dir",
    )
    parser.add_argument(
        "--outputdir-upload",
        help="""Base directory name for uploaded artifacts""",
        dest="outputdir_upload",
    )
    parser.add_argument(
        "--outputdir-local",
        help="""Base directory for locally deployed artifacts, should contain absolut path.""",
        dest="outputdir_local",
    )
    args, _ = parser.parse_known_args()

    if args.outputdir_upload is None or args.outputdir_local is None:
        sys.exit("ERROR: outputdir-local and outputdir-upload needs to be specified.")

    # Get bitbake variables from testdata.json file
    testdata_files = []
    if args.images_dir is not None:
        testdata_files += glob.glob(os.path.join(args.images_dir, "*.testdata.json"))

    with open(testdata_files[0], "r", encoding="utf-8") as f:
        buildvars = json.load(f)

    machine = buildvars["MACHINE"]
    version = buildvars["DISTRO_VERSION"]
    artifacts_all = buildvars["DISTRO_RELEASE_ARTEFACTS"].split()
    artifacts_all.append("BUILD_SRCREVS.log")

    if version.startswith("fngsystem"):
        release_name_local = version.replace("fngsystem", "FNGSystem")
        release_name = release_name_local
    else:
        release_name = "GUF-Yocto-%s" % version
        release_name_local = "Yocto-%s" % version

    output_dir = os.path.join(args.outputdir_upload, release_name)
    outlocal_dir = os.path.join(args.outputdir_local, release_name_local)
    if not os.path.isdir(outlocal_dir):
        sys.exit("ERROR: ouputdir-local does not exist")
    if not os.path.isdir(output_dir):
        sys.exit("ERROR: ouputdir-upload does not exist")

    # Get md5sums.txt
    md5sums_file = os.path.join(output_dir, machine, "md5sums.txt")
    md5sums = {}
    with open(md5sums_file, encoding="utf-8") as f:
        for line in f:
            # Assuming line format: "<md5sum>  <filename>\n"
            name = line.split("  ")[1].rstrip()
            md5sum = line.split("  ")[0]
            md5sums[name] = md5sum

    # Generate alphaplan FWR articles
    generate_fwr_articles(
        output_dir,
        outlocal_dir,
        machine,
        release_name_local,
        artifacts_all,
        md5sums,
    )


if __name__ == "__main__":
    main()
